#!/usr/bin/perl

# bunny - Search in the Debian bugs database
#
# Copyright (C) 2011 Julien Vaubourg <julien@vaubourg.com>
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

use strict;
#use warnings;

use WWW::Mechanize;
use JSON -support_by_pp;
use Getopt::Long;

# URL of the bunny-server
my $bunny_server = "http://bunny.eatmytux.com/bunny-server.pl";

# Encode strings to URL format
sub urlencode {
	my $txt = shift @_;
	$txt =~ s/([^a-z\d])/sprintf("%%%02X", ord($1))/segi;
	return $txt;
}

# Options available
my %opt = (
	"fixed" => undef,
	"sort" => undef,
	"wnpp" => [],
	"tag" => [],
	"severity" => [],
	"max" => 10,
	"id" => [],
	"report" => undef,
	"display" => undef,
	"color" => 0,
	"boolean" => 1,
	"phrase" => 1,
	"lovehate" => 1,
	"anycase" => undef,
	"wildcard" => 1,
	"not" => undef,
	"partial" => undef
);

# Options format
GetOptions(
	'fixed!' => \$opt{"fixed"},
	'use-sort=s' => \$opt{"sort"},
	'wnpp=s' => \@{$opt{"wnpp"}},
	'tag=s' => \@{$opt{"tag"}},
	'severity=s' => \@{$opt{"severity"}},
	'max=i' => \$opt{"max"},
	'id=i' => \@{$opt{"id"}},
	'report=i' => \$opt{"report"},
	'color!' => \$opt{"color"},
	'display=s' => \$opt{"display"},
	'x-boolean!' => \$opt{"boolean"},
	'x-quoted-phrase!' => \$opt{"phrase"},
	'x-lovehate!' => \$opt{"lovehate"},
	'x-any-case-boolean!' => \$opt{"anycase"},
	'x-wildcard!' => \$opt{"wildcard"},
	'x-not!' => \$opt{"not"},
	'x-partial!' => \$opt{"partial"}
);

my $browser = WWW::Mechanize->new();

# With --report option, the user just wants see a single bug report
if($opt{"report"}) {

	eval {
		my $url = "http://bugs.debian.org/cgi-bin/bugreport.cgi?bug=".$opt{"report"};

		# Download the bug report
		$browser->get($url.";mbox=yes");
		my $content = $browser->content();

		# Remove all mail headers except From: To: Subject: and Date:
		$content =~ s/^(From|To|Subject|Date):/[$1]/gm;
		$content =~ s/^[^ :]+:[^\n]*\n(\h+[^\n]+\n)*//gsm;
		$content =~ s/^\n*From [^@]+@[^\n]+\n\[(From|To|Subject|Date)\]/"\n\n".('=' x 80)."\n\n[${1}]"/gsme;

		# Displays the report
		print "BUG #".$opt{"report"};
		print $content;
		print "".('=' x 80)."\n$url\n";

	}; if($@) {
		print STDERR "Bug ID not found.\n";
		exit 2;
	}

# The user wants a research (no -r)
} else {

	my @query;

	# Form the query_string with options (&option=value)
	foreach (keys(%opt)) {
		if($_ !~ /id|fixed|wnpp|tag|severity|display|color/ && $opt{$_}) {
			push(@query, urlencode($_)."=".urlencode($opt{$_}));
		}
	}

	# Booleans filters are added on the Xapian syntax (option:value),
	# directly in the Xapian query, for each value.
	for (qw(id wnpp tag severity)) {
		if(@{$opt{$_}} > 0) {
			for my $value (@{$opt{$_}}) {
				push(@ARGV, $_.":".$value)
			}
		}
	}

	# Fixed is the only boolean filter who can have just a single value
	push(@ARGV, "fixed:".$opt{"fixed"}) if defined $opt{"fixed"};

	# Complete url, with options and Xapian query
	my $url = sprintf(
		$bunny_server."?%s&q=%s&json",
		join('&', @query),
		urlencode(join(' ', @ARGV))
	);

	eval {
		# Send the request
		$browser->get($url);
		my $content = $browser->content();

		# Retrieve the json response
		my $json = new JSON;
		my %result = %{
			$json->allow_nonref->utf8->relaxed->escape_slash->loose->allow_singlequote->allow_barekey->decode($content)
		};

		# If any bugs were found
		if($result{"count"}) {
			my $color = $opt{"color"};

			my %severity_colors = (
				"wishlist" => "\e[0;35m%9s\e[0m",
				"minor" => "\e[0;32m%9s\e[0m",
				"normal" => "%9s",
				"important" => "\e[0;36m%9s\e[0m",
				"serious" => "\e[0;34m%9s\e[0m",
				"grave" => "\e[0;33m%9s\e[0m",
				"critical" => "\e[0;31m%9s\e[0m"
			);

			# Display default value (-d)
			my $display_default = "%pict";

			# Chars to add to the default value
			if($opt{"display"} =~ /^\+/) {
				$opt{"display"} = $display_default.$opt{"display"};

			# Chars to del from the default value
			} elsif($opt{"display"} =~ /^_/) {
				my $del = $opt{"display"};
				$opt{"display"} = $display_default;

				while ($del =~ /(.)/g) {
					$opt{"display"} =~ s/$1//g;
				}
			}

			# If empty : default
			$opt{"display"} = $display_default if $opt{"display"} eq '';

			# Displays the counter header only if -d contains 'c'
			# This is by default
			if($opt{"display"} =~ /c/) {
				print "About " if $result{"about"};
				print $result{"count"}." matching bugs found.\n";
				print "--\n" if length($opt{"display"}) != 1;
			}

			# If the user doesn't want only the counter
			if($opt{"display"} ne 'c') {
				my $date_e = $opt{"display"} =~ /e/;

				$opt{"display"} =~ s/[ce]//g;

				# For each bug found
				for (@{$result{"results"}}) {

					# European date format (-d de)
					if($date_e) {
						$_->{"date"} =~ s/(\d{4})(\d\d)(\d\d)/$3\/$2\/$1/;
					} else {
						$_->{"date"} =~ s/(\d{4})(\d\d)(\d\d)/$1\/$2\/$3/;
					}

					my $percent_color;

					if($color) {
						if($_->{"percent"} > 90) {
							$percent_color = "\e[1;32m%3u%%\e[0m "
						} elsif($_->{"percent"} > 70) {
							$percent_color = "\e[0;36m%3u%%\e[0m "
						} elsif($_->{"percent"} > 50) {
							$percent_color = "\e[0;34m%3u%%\e[0m "
						} elsif($_->{"percent"} > 30) {
							$percent_color = "\e[0;33m%3u%%\e[0m "
						} else {
							$percent_color = "\e[0;31m%3u%%\e[0m "
						}
					}

					# Shows percents (-d %) - default
					printf(
						length($opt{"display"}) != 1 ?
							($color ? $percent_color : "%3u%% - ") : "%u",
						 $_->{"percent"}
					) if $opt{"display"} =~ /%/;

					# Shows dates (-d d)
					printf(
						length($opt{"display"}) != 1 ?
							"[%10s] " : "%s",
						$_->{"date"}
					) if $opt{"display"} =~ /d/;

					# Shows severities (-d s)
					printf(
						length($opt{"display"}) != 1 ?
							($color && $severity_colors{$_->{'severity'}} ?
								"[$severity_colors{$_->{'severity'}}] " :
								"[%9s] "
							) : "%s",
						$_->{"severity"}
					) if $opt{"display"} =~ /s/;

					# Shows fixed flags (-d f)
					printf(
						length($opt{"display"}) != 1 ?
							"[%5s] " : "%s",
						$_->{"fixed"} ? ($color ? "\e[1;32mFIXED\e[0m" : "FIXED") : ""
					) if $opt{"display"} =~ /f/;

					# Shows IDs (-d i) - default
					printf(
						length($opt{"display"}) != 1 ?
							($color ? "[#\e[1m%-7u\e[0m] " : "[#%-7u] ") : "%u",
						$_->{"id"}
					) if $opt{"display"} =~ /i/;

					# Shows package (-d p)
					printf(
						length($opt{"display"}) != 1 ?
							($color ? "\e[4m%s\e[0m " : "(%s) ") : "%s",
						 $_->{"package"}
					) if $opt{"display"} =~ /p/;

					print "-- " if $color && $opt{"display"} =~ /p/ && $opt{"display"} =~ /t/;

					# Shows topics (-d t) - default
					print $_->{"subject"} if $opt{"display"} =~ /t/;

					print "\n";
				}
			}

		# No bug found
		} else {
			exit 1;
		}

	}; if($@) {
		print STDERR "Database access error!\n";
		exit 2;
	}
}

exit 0;

__END__

=head1 NAME

bunny - Search in the Debian bugs database

=head1 SYNOPSIS

bunny [I<OPTIONS>] [--] [I<QUERY>]

=head1 QUERY

QUERY use the Xapian syntax. See the B<EXAMPLES> section.

The double minus (B<-->) can be used for have a QUERY beginning by a minus.

=head2 Free text fields

B<owner:*> - Bug's report submitter

B<doneby:*> - Who fixed the bug

B<package:*> - Package concerned

B<subject:*> - Topic

B<version:*> - Package's version

=head1 OPTIONS

All xapian options and the boolean filter C<fixed> can be negated (e.g.
B<--fixed> and B<--nofixed> or B<--no-fixed>). To negate the others booleans
filters, in a QUERY, use their query version (e.g. C<AND NOT wnpp:ITP>).

=head2 Booleans filters

=over

=item -f, --fixed

Only fixed bugs.

Same as C<fixed:1>.

=item B<[>-id IDB<]...>

Only the bug #ID.

Same as C<id:*>.

=item B<[>-w, --wnpp=I<ITA|ITP|O|RFA|RFH|RFP>B<]...>

Only specifics wnpp bugs.

Same as C<wnpp:*>.

=item B<[>-t, --tag=I<TAG>B<]...>

Only TAG related.

See B<http://www.debian.org/Bugs/Developer#tags>.

Same as C<tag:*>.

=item B<[>-s, --severity=I<wishlist|minor|normal|important|serious|grave|critical>B<]...>

Only specific severity related.

Same as C<severity:*>.

=back

=head2 Format

=over

=item -r, --report=I<ID>

View the bug report #ID.

=item -u, --use-sort=I<date|package|severity|wnpp>

Sort results.

To have reverse order: I<rdate|rpackage|rseverity|rwnpp>

=item -m, --max=I<LIMIT>

Max results. Default 10.

=item -c, --color

Enable bash colors (improves readability).

You can enable it permanently with:

  $ echo 'alias bunny=bunny --color' >> ~/.bashrc
  $ . ~/.bashrc

=item -d, --display=I<DISPLAY>

B<c> - Shows counter header

B<d> - Shows dates

B<e> - Shows date with DD/MM/YYYY format

B<f> - Shows the fixed tag

B<i> - Shows IDs

B<p> - Shows percents

B<s> - Shows severity tags

B<t> - Shows topics

B<Default:> C<%pict>

B<Examples:> C<-d fist%depc>,  C<-d +sd>,  C<-d _p>

You could use a bash alias to redefine C<--display> permanently (see above).

=back

=head2 Xapian options

=over

=item -x-b, --x-boolean

Support C<AND>, C<OR>, etc and bracketed subexpressions.

=item -x-q, --x-quoted-phrase B<(default)>

Support quoted phrases.

=item -x-l, --x-lovehate B<(default)>

Support C<+> and C<->.

=item -x-a, --x-any-case-boolean

Support C<AND>, C<OR>, etc even if they aren't in ALLCAPS.

=item -x-w, --x-wildcard B<(default)>

Support right truncation (not on B<Booleans filters>).

=item -x-n, --x-not

Allow queries such as C<NOT debian>.

These require the use of a list of all bugs in the database which is expensive,
so this feature isn't enabled by default.

=item -x-p, --x-partial

Enable partial matching.

Partial matching causes the parser to treat the query as a "partially entered"
search. This will automatically treat the final word as a wildcarded match,
unless it is followed by whitespace, to produce more stable results from
interactive searches.

Currently B<--x-partial> doesn't do anything if the final word in the query has
a boolean filter prefix, or if it is in a phrase (either an explicitly quoted
one, or one implicitly generated by hyphens or other punctuation). It also
doesn't do anything if if the final word is part of a value range.

=back

=head1 BASH AUTO COMPLETION

Suggest values for tab completion of the current command line. Works with all
the options and several free text fields & boolean filters.

You must have bash-completion installed and enabled:

  # apt-get install bash-completion
  $ echo '[ -f /etc/bash_completion ] && . /etc/bash_completion' >> ~/.bash_profile
  $ . ~/.bash_profile

=head1 EXAMPLES

 bunny owner:john severity:grave tag:sid version:3.5.13
 bunny -x-n --nofixed package:iceweasel AND NOT owner:john
 bunny doneby:'john doe' 99/12/25..01/12/31
 bunny package:'iceweasel*' it\'s embarassing
 bunny -w ITP -w RFP -w O
 bunny -r 421337 | less
 bunny -d +s package:gnome --color

=head1 EXIT STATUS

B<0> - Matching bugs found

B<1> - No matching bugs found

B<Others> - Error

=head1 AUTHOR

Julien Vaubourg E<lt>julien@vaubourg.comE<gt>

=head1 COPYRIGHT

Copyright (C) 2011 Julien Vaubourg <julien@vaubourg.com>

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation, either version 3 of the License, or later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program. If not, see <http://www.gnu.org/licenses/>. On Debian systems, the
complete text of version 3 of the GNU General Public License can be found in
</usr/share/common-licenses/GPL-3>.

=cut
